--Autoren:
--Tobias Möllmann, inf9407
--Felix-Alexander Döppers inf7340

import List
    ( (\\)
    )

import System
    ( getArgs
    )

-- data types for the sudoku solver

-- matrix
type Matrix a = [Row a]
type Row a    = [a]

-- 9 x 9 grid with Digits

type Grid    = Matrix Digit
type Digit   = Char

type Choices = [Digit]

-- digits: ['1'-'9'] and '0' as blank
digits :: [Digit]
digits = ['1'..'9']

blank :: Digit -> Bool
blank  = (== '0')

-- solver algorithm
solve :: Grid -> [Grid]
solve = filter valid . expand . prune . choices

prune :: Matrix Choices -> Matrix Choices
prune = pruneBy boxs . pruneBy cols . pruneBy rows
        where pruneBy f = f . map pruneRow . f

pruneRow :: Row Choices -> Row Choices
pruneRow row = map (remove fixed) row
               where fixed = [d | [d] <- row]

isSingleton :: [a] -> Bool
isSingleton [_] = True
isSingleton _ = False

remove xs ds = if isSingleton ds then ds else ds \\ xs

choices :: Grid -> Matrix Choices
choices = map (map choice)
  where choice d = if blank d then digits else [d]

expand :: Matrix Choices -> [Grid]
expand = cp . map cp


--cartesian product
cp            :: [[a]] -> [[a]]
cp []         = [[]]
cp (xs : xss) = [x : ys | x <- xs, ys <- cp xss]

valid   :: Grid -> Bool
valid g = all nodups (rows g) &&
          all nodups (cols g) &&
          all nodups (boxs g)

nodups          :: Eq a => [a] -> Bool
nodups []       = True
nodups (x : xs) = all (/= x) xs && nodups xs

rows :: Matrix a -> Matrix a
rows = id

cols            :: Matrix a -> Matrix a
cols [xs]       = [[x]| x <- xs]
cols (xs : xss) = zipWith (:) xs (cols xss)

boxs :: Matrix a -> Matrix a
boxs = map ungroup . ungroup . map cols . group . map group

group    :: [a] -> [[a]]
group [] = []
group xs = take 3 xs : group (drop 3 xs)

ungroup :: [[a]] -> [a]
ungroup = concat

readSudoku :: String -> Grid
readSudoku xs = toCols (suRows xs)
              where
              toCols (xs : xss) = (suCols xs) : (toCols xss)
              toCols [] = []
              suRows xs = snd $ unzip $ zip [1..9] (lines xs)
              suCols ys = map head (snd $ unzip $ zip [1..9] (words ys))

showSudoku     :: Grid -> IO()
showSudoku xss = putStr (unlines (insSpaces xss))

insSpaces :: [String] -> [String]
insSpaces (xs : xss) = (unwords (map (\x -> [x]) xs)) : ( insSpaces xss)
insSpaces []       = []

main :: IO ()
main = do
      [f] <- getArgs
      xs <- readFile f
      showSudoku (head (solve (readSudoku xs)))
